package pascal.semantic;

import pascal.analysis.*;
import pascal.node.*;
import pascal.table.*;
import pascal.iterator.*;

import java.util.Hashtable;

public class SemanticAnalyzer extends DepthFirstAdapter
{
  private SymbolTable table;
  private Hashtable info;
  private Type type_int, type_real, type_char, type_bool, type_string;
  private ConcreteAggregate error_list;

  public SemanticAnalyzer(){
    TableFactory factory = TableFactory.getInstance();
    table = factory.getSymbolTable();
    info = new Hashtable();
    error_list = new ConcreteAggregate();
  }

  public void inAProgram(AProgram node){
    table.enterBlock();
    try{
      type_int=new Type("int");
      table.enter(type_int);

      type_real=new Type("real");
      table.enter(type_real);

      type_char=new Type("char");
      table.enter(type_char);

      type_bool=new Type("bool");
      table.enter(type_bool);

      type_string=new Type("string");
      table.enter(type_string);

    }catch(CompileError e){
      error_list.add(e);
    }
  }
  
  public void outAProgram(AProgram node){
    table.exitBlock();
    if(error_list.getLength() > 0){
      Iterator it = error_list.iterator();
      while(it.hasNext()){
        System.out.println(((CompileError)it.next()).getMessage());
      }
      System.exit(1);
    }
  }

  public void outAProgramHead(AProgramHead node){
    Information ident = (Information)info.get(node.getIdent());
    Information file_name_list = (Information)info.get(node.getFileNameList());
    try{
      table.enter(new Program(ident.getString(), file_name_list.getList()));
      table.enterBlock();
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }
  }

  public void outAInputFileNameList(AInputFileNameList node){
    Information file_name_list = new Information();
    ConcreteAggregate list = new ConcreteAggregate();
    list.add(node.getInput());
    list.add(node.getOutput());
    file_name_list.set(list);
    info.put(node, file_name_list);
  }  

  public void outAOutputFileNameList(AOutputFileNameList node){
    Information file_name_list = new Information();
    ConcreteAggregate list = new ConcreteAggregate();
    list.add(node.getOutput());
    file_name_list.set(list);
    info.put(node, file_name_list);
  }  

  public void outABlock(ABlock node){
    table.exitBlock();
  }

  public void outAConstDecl(AConstDecl node){
    Information ident = (Information)info.get(node.getIdent());
    Information constant = (Information)info.get(node.getConstant());
    try{
      if(constant.getType() == type_string){
        error_list.add(new SemanticError(ident.getLine(), 
                                         ident.getPos(),
                                         "error!!"));
      }else{
        table.enter(new Constant(ident.getString(),
                                 constant.getType(),
                                 constant.getString()));
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }
  }

  public void outATypeDecl(ATypeDecl node){
    Information ident = (Information)info.get(node.getIdent());
    Information type = (Information)info.get(node.getType());
    try{
      table.enter(new TypeIdent(ident.getString(), type.getType()));
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }
  }

  public void outAVarDecl(AVarDecl node){
    Information ident_list = (Information)info.get(node.getIdentList());
    Information type = (Information)info.get(node.getType());
    ConcreteAggregate list = ident_list.getList();
    Iterator it = list.iterator();
    while(it.hasNext()){
      Information ident = (Information)it.next();
      try{
        table.enter(new Variable(ident.getString(), type.getType()));
      }catch(CompileError e){
        e.setLinePos(ident.getLine(), ident.getPos());
        error_list.add(e);
      }
    }
  }


  public void outAIdentListIdentList(AIdentListIdentList node){
    Information ident_list = (Information)info.get(node.getIdentList());
    ConcreteAggregate list = ident_list.getList();
    list.add(info.get(node.getIdent()));
    info.put(node, ident_list);
  }

  public void outAIdentIdentList(AIdentIdentList node){
    Information ident_list = new Information();
    ConcreteAggregate list = new ConcreteAggregate();
    list.add(info.get(node.getIdent()));
    ident_list.set(list);
    info.put(node, ident_list);
  }

  public void outAIdent(AIdent node){
    Information ident = new Information();
    ident.set(node.getIdentt().getText().trim());
    ident.set(node.getIdentt().getLine(), node.getIdentt().getPos());
    info.put(node, ident);
  }

  public void outAProcDeclHead(AProcDeclHead node){
    Information ident = (Information)info.get(node.getIdent());
    Information formal_param_list
               = (Information)info.get(node.getFormalParamList());
    ConcreteAggregate list = new ConcreteAggregate();
    ConcreteAggregate list2 = formal_param_list.getList();
    try{
      Procedure procedure = new Procedure(ident.getString(), list);
      table.enter(procedure);

      table.enterBlock();
      if(list2 != null){
        Iterator it2 = list2.iterator();
        while(it2.hasNext()){
          Information formal_param_decl = (Information)it2.next();
          ConcreteAggregate list3 = formal_param_decl.getList();
          Iterator it3 = list3.iterator();
          while(it3.hasNext()){
            Information formal_param = (Information)it3.next(); 
            try{
              table.enter(formal_param.getParameter());
              list.add(formal_param.getParameter());
            }catch(CompileError e){
              e.setLinePos(formal_param.getLine(), formal_param.getPos());
              error_list.add(e);
            }
          }
        }
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }
  }

  public void outAFuncDeclHead(AFuncDeclHead node){
    Information ident = (Information)info.get(node.getIdent());
    Information formal_param_list = (Information)info.get(node.getFormalParamList());
    Information type = (Information)info.get(node.getTypeIdent());
    ConcreteAggregate list = new ConcreteAggregate();
    ConcreteAggregate list2 = formal_param_list.getList();

    try{
      if(type.getType() instanceof ArrayType || type.getType() instanceof RecordType){
        error_list.add(new SemanticError(type.getLine(),
                                            type.getPos(),
                                            "error!!(102)"));
      }      
      Function function = new Function(ident.getString(), list, type.getType());
      table.enter(function);

      table.enterBlock();
      if(list2 != null){
        Iterator it2 = list2.iterator();
        while(it2.hasNext()){
          Information formal_param_decl = (Information)it2.next();
          ConcreteAggregate list3 = formal_param_decl.getList();
          Iterator it3 = list3.iterator();
          while(it3.hasNext()){
            Information formal_param = (Information)it3.next(); 
            try{
              table.enter(formal_param.getParameter());
              list.add(formal_param.getParameter());
            }catch(CompileError e){
              e.setLinePos(formal_param.getLine(), formal_param.getPos());
              error_list.add(e);
            }
          }
        }
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }
  }

  public void outAStdTypeTypeIdent(AStdTypeTypeIdent node){
    info.put(node, info.get(node.getStdType()));
  }

  public void outAIdentTypeIdent(AIdentTypeIdent node){
    Information type = new Information();
    Information ident = (Information)info.get(node.getIdent());
    try{
      Entry entry = table.lookup(ident.getString());
      if(!(entry instanceof TypeIdent)){
        error_list.add(new SemanticError(ident.getLine(),
                                            ident.getPos(),
                                            ident + " does not be declared as TYPE"));
      }else{
        type.set(((TypeIdent)entry).getType());
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }finally{
      type.set(ident.getLine(), ident.getPos());
      info.put(node, type);
    }
  }

  public void outATypeIntegerStdType(ATypeIntegerStdType node){
    Information type = new Information();
    type.set(type_int);
    info.put(node, type);
  }

  public void outATypeRealStdType(ATypeRealStdType node){
    Information type = new Information();
    type.set(type_real);
    info.put(node, type);
  }

  public void outATypeCharStdType(ATypeCharStdType node){
    Information type = new Information();
    type.set(type_char);
    info.put(node, type);
  }

  public void outATypeBoolStdType(ATypeBoolStdType node){
    Information type = new Information();
    type.set(type_bool);
    info.put(node, type);
  }

  public void outATypeType(ATypeType node){
    info.put(node, info.get(node.getTypeIdent()));
  }

  public void outAArrayType(AArrayType node){
    Information array_type = new Information();
    ConcreteAggregate list = new ConcreteAggregate();
    Information type = (Information)info.get(node.getType());
    Information index_list = (Information)info.get(node.getIndexList());
    ConcreteAggregate list2 = index_list.getList();
    Iterator it = list2.iterator();
    while(it.hasNext()){
      Information index = (Information)it.next();
      list.add(index.getEntry());
    }
    ArrayType type2 = new ArrayType(node.getArray().getLine(), node.getArray().getPos(),
                                    list, type.getType());
    array_type.set(type2);
    try{
      table.enter(type2);
    }catch(CompileError e){
      e.setLinePos(node.getArray().getLine(), node.getArray().getPos());
      error_list.add(e);
    }finally{
      info.put(node, array_type);
    }
  }

  public void inARecordType(ARecordType node){
    Information record_type = new Information();
    RecordType type = new RecordType(node.getRecord().getLine(), node.getRecord().getPos());
    record_type.set(type);
    try{
      table.enter(type);
      table.enterRecord();
    }catch(CompileError e){
      e.setLinePos(node.getRecord().getLine(), node.getRecord().getPos());
      error_list.add(e);
    }finally{
      info.put(node, record_type);
    }
  }

  public void outARecordType(ARecordType node){
    Information record_type = (Information)info.get(node);
    RecordType type = (RecordType)record_type.getType();
    int i = table.exitRecord();
    type.setField(i);
  }

  public void outAIndexListIndexList(AIndexListIndexList node){
    Information index_list = (Information)info.get(node.getIndexList());
    ConcreteAggregate list = index_list.getList();
    list.add(info.get(node.getIndex()));
    info.put(node, index_list);
  }

  public void outAIndexIndexList(AIndexIndexList node){
    Information index_list = new Information();
    ConcreteAggregate list = new ConcreteAggregate();
    list.add(info.get(node.getIndex()));
    index_list.set(list);
    info.put(node, index_list);
  }

  public void outAIndex(AIndex node){
    Information index = new Information();
    Information left = (Information)info.get(node.getLeft());
    Information right = (Information)info.get(node.getRight());

    Type left_type = left.getType();
    Type right_type = right.getType();
    if(!((left_type == type_int && right_type == type_int)
	 || (left_type == type_char && right_type == type_char))){
      error_list.add(new SemanticError(left.getLine(), 
                                          left.getPos(),
                                          "Index is not scaler type"));
    }else{
      Index left_right = new Index(left_type,
                                   left.getString(),
                                   right.getString());
      int range = left_right.getRange(); 
      if(range < 1){
        error_list.add(new SemanticError(left.getLine(),
                                            left.getPos(), 
                                            "Index is not scaler"));
      }else{
        index.set(left_right);
      }
    }
    info.put(node, index);
  }

  public void outARecordBody(ARecordBody node){
    Information ident_list = (Information)info.get(node.getIdentList());
    Information type = (Information)info.get(node.getType());
    ConcreteAggregate list = ident_list.getList();
    Type field_type = type.getType();

    Iterator it = list.iterator();
    while(it.hasNext()){
      Information ident = (Information)it.next();
      try{
        table.enter(new Field(ident.getString(), field_type));
      }catch(CompileError e){
        e.setLinePos(ident.getLine(), ident.getPos());
        error_list.add(e);
      }
    }
  }

  public void outAFormalParamDeclListFormalParamList
                      (AFormalParamDeclListFormalParamList node){
    info.put(node, info.get(node.getFormalParamDeclList()));
  }

  public void outAEmptyFormalParamList(AEmptyFormalParamList node){
    info.put(node, new Information());
  }

  public void outAFormalParamDeclListFormalParamDeclList
                                 (AFormalParamDeclListFormalParamDeclList node){
    Information formal_param_decl_list 
                   = (Information)info.get(node.getFormalParamDeclList());
    ConcreteAggregate list = formal_param_decl_list.getList();
    list.add(info.get(node.getFormalParamDecl()));
    info.put(node, formal_param_decl_list);
  }

  public void outAFormalParamDeclFormalParamDeclList
                                      (AFormalParamDeclFormalParamDeclList node){
    Information param_list = new Information();
    ConcreteAggregate list = new ConcreteAggregate();
    list.add(info.get(node.getFormalParamDecl()));
    param_list.set(list);
    info.put(node, param_list);
  }

  public void outAVarFormalParamDecl(AVarFormalParamDecl node){
    Information param_decl = new Information();
    ConcreteAggregate formal_param_list = new ConcreteAggregate();
    Information decl_ident_list 
               = (Information)info.get(node.getFormalParamDeclIdentList());
    Type type = decl_ident_list.getType();
    ConcreteAggregate ident_list = decl_ident_list.getList();
    Iterator it = ident_list.iterator();
    while(it.hasNext()){
      Information ident = (Information)it.next();
      Information param = new Information();
      param.set(new VariableParameter(ident.getString(), type));
      param.set(ident.getLine(), ident.getPos());
      formal_param_list.add(param);        
    }
    param_decl.set(formal_param_list);
    info.put(node, param_decl);
  }

  public void outAFormalParamDeclIdentListFormalParamDecl
                               (AFormalParamDeclIdentListFormalParamDecl node){
    Information param_decl = new Information();
    ConcreteAggregate formal_param_list = new ConcreteAggregate();
    Information decl_ident_list 
                = (Information)info.get(node.getFormalParamDeclIdentList());
    Type type = decl_ident_list.getType();
    ConcreteAggregate ident_list = decl_ident_list.getList();
    Iterator it = ident_list.iterator();
    while(it.hasNext()){
      Information ident = (Information)it.next();
      Information param = new Information();
      param.set(new ValueParameter(ident.getString(), type));
      param.set(ident.getLine(), ident.getPos());
      formal_param_list.add(param);  
    }
    param_decl.set(formal_param_list);
    info.put(node, param_decl);
  }

  public void outAFormalParamDeclIdentList(AFormalParamDeclIdentList node){
    Information decl_ident_list = new Information(); 
    Information ident_list = (Information)info.get(node.getIdentList());
    Information type_ident = (Information)info.get(node.getTypeIdent());
    decl_ident_list.set(ident_list.getList());
    decl_ident_list.set(type_ident.getType());
    info.put(node, decl_ident_list);
  }

  public void outAVariableTailAssignment(AVariableTailAssignment node){
    Information ident = (Information)info.get(node.getIdent());
    Information variable_tail = (Information)info.get(node.getVariableTail());
    Information expr = (Information)info.get(node.getExpr()); 
    try{
      Type left_type = getVariableTailTypeL(ident,
                                            variable_tail.getList(),
                                            table.getCurrent());
      Type right_type = expr.getType();
      if(!((left_type == right_type)
            || (left_type==type_real && right_type==type_int))){
        error_list.add(new SemanticError(ident.getLine(),
                                            ident.getPos(),
                                            "Mismatch type in assignment statement"));
      } 
    }catch(Exception e){
      error_list.add(e);
    }
  }

  public void outAColeqAssignment(AColeqAssignment node){
     Information ident = (Information)info.get(node.getIdent());
     Information expr = (Information)info.get(node.getExpr());
     try{
     Entry entry = table.lookup(ident.getString());
      if(!(entry instanceof Variable || entry instanceof Function 
          ||entry instanceof Parameter)){
            error_list.add(new SemanticError(ident.getLine(),
                                                ident.getPos(),
                                                ident.getString() + " can not assign"));
      }else{
        if(entry instanceof Function && entry != table.getCurrent()){
            error_list.add(new SemanticError(ident.getLine(),
                                                ident.getPos(),
                                                ident.getString() + " can not assign"));
        }
        Type left_type = ((hasType)entry).getType();
        Type right_type = expr.getType();
        if(!((left_type == right_type)
              || (left_type==type_real && right_type==type_int))){
            error_list.add( new SemanticError(node.getColeq().getLine(),
                                                 node.getColeq().getPos(), 
                                                "Mismatch type in assignment statement"));
        } 
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }
  }

  public void outAActualParamListProcCall(AActualParamListProcCall node){
    Information ident = (Information)info.get(node.getIdent());
    Information actual_param_list = (Information)info.get(node.getActualParamList());

    try{
      Entry entry = table.lookup(ident.getString());      
      if(!(entry instanceof Procedure)){
        error_list.add(new SemanticError(ident.getLine(),
                                            ident.getPos(), 
                                            ident +" is not Procedure"));
      }

      ConcreteAggregate actual_param = actual_param_list.getList();
      if(((Procedure)entry).getNumberOfParameter() != actual_param.getLength()){
        error_list.add(new SemanticError(ident.getLine(),
                                            ident.getPos(), 
                                            "Not match parameter"));
      }
      Iterator actual_param_it = actual_param.iterator();
      Iterator formal_param_it = ((Procedure)entry).getParameter();
      while(actual_param_it.hasNext()){
        Parameter formal = (Parameter)formal_param_it.next();
        Information actual = (Information)actual_param_it.next();
        if(formal.getType() != actual.getType()){
          error_list.add(new SemanticError(ident.getLine(),
                                              ident.getPos(),
                                              "Not match parameter"));
        }
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }
  }

  public void outAIdentProcCall(AIdentProcCall node){
    Information ident = (Information)info.get(node.getIdent());
    try{
      Entry entry = table.lookup(ident.getString());
      if(!(entry instanceof Procedure)){
          error_list.add(new SemanticError(ident.getLine(),
                                              ident.getPos(),
                                              ident+" is not Procedure"));
      }

      if(((Procedure)entry).getNumberOfParameter() != 0){
        error_list.add(new SemanticError(ident.getLine(),
                                            ident.getPos(),
                                            "Not match parameter"));
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }
  }

  public void outAReadBuiltinProcCall(AReadBuiltinProcCall node){
    Information actual_param_list = (Information)info.get(node.getActualParamList());
    ConcreteAggregate list = actual_param_list.getList();
    Iterator it = list.iterator();
    while(it.hasNext()){
      Information expr = (Information)it.next();
      if(!expr.isRVariable()){
          error_list.add(new SemanticError(expr.getLine(),
                                              expr.getPos(),
                                              "Not match parameter"));
      }
    } 
  }

  public void outAReadlnBuiltinProcCall(AReadlnBuiltinProcCall node){
    Information actual_param_list = (Information)info.get(node.getActualParamList());
    ConcreteAggregate list = actual_param_list.getList();
    Iterator it = list.iterator();
    while(it.hasNext()){
      Information expr = (Information)it.next();
      if(!expr.isRVariable()){
          error_list.add(new SemanticError(expr.getLine(),
                                              expr.getPos(),
                                              "Not match parameter"));
      }
    } 
  }

  public void outAWriteBuiltinProcCall(AWriteBuiltinProcCall node){
    Information actual_param_list = (Information)info.get(node.getActualParamList());
    ConcreteAggregate list = actual_param_list.getList();
    Iterator it = list.iterator();
    while(it.hasNext()){
      Information expr = (Information)it.next();
      Type expr_type = expr.getType();
      if(expr_type == type_bool){
        error_list.add(new SemanticError(expr.getLine(), 
                                            expr.getPos(),
                                            "Not match parameter"));
      }
    }
  }

  public void outAWritelnlparBuiltinProcCall(AWritelnlparBuiltinProcCall node){
    Information actual_param_list = (Information)info.get(node.getActualParamList());
    ConcreteAggregate list = actual_param_list.getList();
    Iterator it = list.iterator();
    while(it.hasNext()){
      Information expr = (Information)it.next();
      Type expr_type = expr.getType();
      if(expr_type == type_bool){
        error_list.add(new SemanticError(expr.getLine(),
                                            expr.getPos(),
                                            "Not match parameter"));
      }
    }
  }

  public void outAFmatchedStmt(AFmatchedStmt node){
    Information expr = (Information)info.get(node.getExpr());
    Type expr_type = expr.getType();
    if(expr_type !=type_bool){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(),
                                          "Expression(IF-CONDITION) is not boolean type"));
    }
  }

  public void outAIfthenIfUnmatchedStmt(AIfthenIfUnmatchedStmt node){
    Information expr = (Information)info.get(node.getExpr());
    Type expr_type = expr.getType();
    if(expr_type !=type_bool){
      error_list.add(new SemanticError(expr.getLine(), 
                                          expr.getPos(),
                                          "Expression(IF-CONDITION) is not boolean type"));
    }
  }

  public void outAIfthenUnmatchedStmt(AIfthenUnmatchedStmt node){
    Information expr = (Information)info.get(node.getExpr());
    Type expr_type = expr.getType();
    if(expr_type !=type_bool){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(), 
                                          "Expression(IF-CONDITION) is not boolean type"));
    }
  }

  public void outAIfthenelseUnmatchedStmt(AIfthenelseUnmatchedStmt node){
    Information expr = (Information)info.get(node.getExpr());
    Type expr_type = expr.getType();
    if(expr_type !=type_bool){
      error_list.add(new SemanticError(expr.getLine(), 
                                          expr.getPos(), 
                                          "Expression(IF-CONDITION) is not boolean type"));
    }
  }

  public void outACaseStmt(ACaseStmt node){
    TableFactory factory = TableFactory.getInstance();
    CaseLabelTable case_label_table 
           = factory.getCaseLabelTable(node.getCase().getLine(), 
                                       node.getCase().getPos());

    Information expr = (Information)info.get(node.getExpr());
    Information case_body_list = (Information)info.get(node.getCaseBodyList());

    Type expr_type = expr.getType();
    if(!(expr_type == type_char || expr_type == type_int)){
      error_list.add(new SemanticError(expr.getLine(), 
                                          expr.getPos(),
                                          "Expression(Case) is not scaler"));
    }else{
      ConcreteAggregate list = case_body_list.getList();
      Iterator it = list.iterator();
      while(it.hasNext()){
        Information case_body = (Information)it.next();
        ConcreteAggregate list2 = case_body.getList();
        if(list2 != null){
          Iterator it2 = list2.iterator();
          while(it2.hasNext()){
            Information case_label = (Information)it2.next();
            Type case_label_type = case_label.getType();
            String case_label_name = case_label.getString();
            if(expr_type != case_label_type){
              error_list.add(new SemanticError(case_label.getLine(),
                                                  case_label.getPos(),
                                                  "Case Labels' type mismatch case expression"));
            }else{
              try{
                case_label_table.enter(new CaseLabel(case_label_name));
              }catch(CompileError e){
                e.setLinePos(case_label.getLine(), case_label.getPos());
                error_list.add(e);
              }
            }  
          }
        }
      }
    }
  }

  public void outACaseBodyListCaseBodyList(ACaseBodyListCaseBodyList node){
    Information case_body_list = (Information)info.get(node.getCaseBodyList());
    ConcreteAggregate list = case_body_list.getList();
    list.add(info.get(node.getCaseBody()));
    info.put(node, case_body_list);
  }

  public void outACaseBodyCaseBodyList(ACaseBodyCaseBodyList node){
    Information body_list = new Information();
    ConcreteAggregate list = new ConcreteAggregate();
    list.add(info.get(node.getCaseBody()));
    body_list.set(list);
    info.put(node, body_list);
  }

  public void outACaseLabelListCaseBody(ACaseLabelListCaseBody node){
    info.put(node, info.get(node.getCaseLabelList()));
  }

  public void outAEmptyCaseBody(AEmptyCaseBody node){
    info.put(node, new Information());
  }

  public void outACaseLabelListCaseLabelList(ACaseLabelListCaseLabelList node){
    Information case_label_list = (Information)info.get(node.getCaseLabelList());
    ConcreteAggregate list = case_label_list.getList();
    list.add(info.get(node.getCaseLabel()));
    info.put(node, case_label_list);
  }

  public void outACaseLabelCaseLabelList(ACaseLabelCaseLabelList node){
    Information label_list = new Information();
    ConcreteAggregate list = new ConcreteAggregate();
    list.add(info.get(node.getCaseLabel()));
    label_list.set(list);
    info.put(node, label_list);
  }

  public void outACaseLabel(ACaseLabel node){
    info.put(node, info.get(node.getConstant()));
  }

  public void outAWhileStmt(AWhileStmt node){
    Information expr = (Information)info.get(node.getExpr());
    Type expr_type = expr.getType();
    if(expr_type != type_bool){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(), 
                                          "Expression(While) is not boolean type"));
    }
  }

  public void outARepeatStmt(ARepeatStmt node){
    Information expr = (Information)info.get(node.getExpr());
    Type expr_type = expr.getType();
    if(expr_type != type_bool){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(),
                                          "Expression(REPEAT-UNTIL) is not boolean type"));
    }
  }

  public void outAForStmt(AForStmt node){
    Information ident = (Information)info.get(node.getIdent());
    Information left = (Information)info.get(node.getLeft());
    Information right = (Information)info.get(node.getRight());

    try{
      Entry entry = table.lookup(ident.getString());
      if(!(entry instanceof Variable || entry instanceof Parameter)){
        error_list.add(new SemanticError(ident.getLine(),
                                            ident.getPos(), 
                                            ident.getString() + "is not variable"));
      }else{
        Type type=((hasType)entry).getType();
        if(type != type_int){
          error_list.add(new SemanticError(ident.getLine(),
                                              ident.getPos(), 
                                              ident.getString() + " is not integer"));
        } 

        Type left_type = left.getType();
        Type right_type = right.getType();
        if(left_type != type_int){
         error_list.add(new SemanticError(left.getLine(),
                                             left.getPos(), 
                                             "Use start-range integer"));
        }else{
          if(right_type != type_int){
            error_list.add(new SemanticError(right.getLine(),
                                                right.getPos(),
                                                "Use end-range integer"));
          }
        }
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }
  }

  public void outASimpleExprExpr(ASimpleExprExpr node){
    info.put(node, info.get(node.getSimpleExpr()));
  }

  public void outASimpleExprrelopsimpleExprExpr(ASimpleExprrelopsimpleExprExpr node){
    Information expr = new Information();
    Information left = (Information)info.get(node.getLeft());
    Information right = (Information)info.get(node.getRight());

    Type left_type = left.getType();
    Type right_type = right.getType();
    if(!(left_type == type_int || left_type == type_real)){
      error_list.add(new SemanticError(left.getLine(),
                                          left.getPos(),
                                          "Left hand side expression is not Integer/Real"));
    }else{
      if(!(right_type == type_int || right_type == type_real)){
        error_list.add(new SemanticError(right.getLine(),
                                            right.getPos(),
                                            "Right hand side expression is not Integer/Real"));
      }else{
        expr.set(type_bool);
        expr.setRValue();
      }
    }
    expr.set(left.getLine(), left.getPos());
    info.put(node, expr);
  }

  public void outATermSimpleExpr(ATermSimpleExpr node){
    info.put(node, info.get(node.getTerm()));
  }

  public void outAUnoptermSimpleExpr(AUnoptermSimpleExpr node){
    Information simple_expr = new Information();
    Information term = (Information)info.get(node.getTerm());
    Type term_type = term.getType();
    if(!(term_type == type_int || term_type == type_real)){
      error_list.add(new SemanticError(term.getLine(),
                                          term.getPos(),
                                          "Unary operand is not integer/real"));
    }else{
      simple_expr = term;
      simple_expr.setRValue();
    }
    info.put(node, simple_expr);
  }

  public void outASimpleExprlowOptermSimpleExpr(ASimpleExprlowOptermSimpleExpr node){
    Information simple_expr = new Information();
    Information simple_expr2 = (Information)info.get(node.getSimpleExpr());
    Information term = (Information)info.get(node.getTerm());

    Type simple_expr_type = simple_expr2.getType();
    Type term_type = term.getType();
    if(!(simple_expr_type == type_int || simple_expr_type == type_real)){
      error_list.add(new SemanticError(simple_expr2.getLine(),
                                          simple_expr2.getPos(),
                                          "Left hand Operand is not Integer/Real"));
    }else{
      if(!(term_type == type_int || term_type == type_real)){
          error_list.add(new SemanticError(term.getLine(),
                                              term.getPos(),
                                              "Right hand operand is not Integer/Real"));
      }else{
        if(simple_expr_type == type_int &&  term_type==type_int){
          simple_expr.set(type_int);
          simple_expr.setRValue();
        }else{
          simple_expr.set(type_real);
          simple_expr.setRValue();
        }
      }
    }
    simple_expr.set(simple_expr2.getLine(), simple_expr2.getPos());
    info.put(node, simple_expr);
  }

  public void outASimpleExprorOptermSimpleExpr(ASimpleExprorOptermSimpleExpr node){
    Information simple_expr = new Information();
    Information simple_expr2 = (Information)info.get(node.getSimpleExpr());
    Information term = (Information)info.get(node.getTerm());

    Type simple_expr_type = simple_expr2.getType();
    Type term_type = term.getType();
    if(simple_expr_type != type_bool){
      error_list.add(new SemanticError(simple_expr2.getLine(),
                                          simple_expr2.getPos(),
                                          "Left hand side operand is not Boolean"));
    }else{
      if(term_type != type_bool){
        error_list.add(new SemanticError(term.getLine(),
                                            term.getPos(), 
                                            "Right hand operand is not Boolean"));
      }else{
        simple_expr.set(type_bool);
        simple_expr.setRValue();
      }
    }
    simple_expr.set(simple_expr2.getLine(),simple_expr2.getPos());
    info.put(node, simple_expr);
  }

  public void outAFactTerm(AFactTerm node){
    info.put(node, info.get(node.getFact()));
  }

  public void outATermhighOpfactTerm(ATermhighOpfactTerm node){
    Information term = new Information();
    Information term2 = (Information)info.get(node.getTerm());
    Information fact = (Information)info.get(node.getFact());

    Type term_type = term2.getType();
    Type fact_type = fact.getType();
    if(node.getHighOp() instanceof AMultHighOp){ //|Z(*)
      if(!(term_type == type_int || term_type == type_real)){
        error_list.add(new SemanticError(term2.getLine(), 
                                            term2.getPos(),
                                            "Left hand side operand is not Integer/Real"));
      }else{
        if(!(fact_type == type_int || fact_type == type_real)){
          error_list.add(new SemanticError(fact.getLine(), 
                                              fact.getPos(),
                                              "Right hand side operand is not Integer/Real"));
        }else{
          if(fact_type == type_int && term_type == type_int){
            term.set(type_int);
            term.setRValue();
          }else{
            term.set(type_real);
          }
        }
      }
    }else if(node.getHighOp() instanceof ASlashHighOp){ // (/)
      if(!(term_type == type_int || term_type == type_real)){
        error_list.add(new SemanticError(term2.getLine(),
                                            term2.getPos(),
                                            "Left hand side operand is not Integer/Real"));
      }else{
        if(!(fact_type == type_int || fact_type == type_real)){
          error_list.add(new SemanticError(fact.getLine(), 
                                              fact.getPos(),
                                              "Right hand side operand is not Integer/Real"));
        }else{
          term.set(type_real);
          term.setRValue();
        }
      }
    }else if(node.getHighOp() instanceof ADivHighOp 
             || node.getHighOp() instanceof AModHighOp){
      if(term_type != type_int){
         error_list.add(new SemanticError(term2.getLine(),
                                             term2.getPos(),
                                             "Left hand side operand is not Integer"));
      }else{
        if(fact_type != type_int){
          error_list.add(new SemanticError(fact.getLine(),
                                              fact.getPos(),
                                              "Right hand side operand is not Integer"));
        }else{
          term.set(type_int);
          term.setRValue();
        }
      }
    }
    term.set(term2.getLine(), term2.getPos());
    info.put(node, term);
  }

  public void outATermandOpfactTerm(ATermandOpfactTerm node){
    Information term = new Information();
    Information term2 = (Information)info.get(node.getTerm());
    Information fact = (Information)info.get(node.getFact());

    Type term_type = term2.getType();
    Type fact_type = fact.getType();
    if(term_type != type_bool){
      error_list.add(new SemanticError(term2.getLine(),
                                          term2.getPos(),
                                          "Left hand side operand is not Integer/Real"));
    }else{
      if(fact_type != type_bool){
        error_list.add(new SemanticError(fact.getLine(), 
                                            fact.getPos(),
                                            "Right hand side operand is not Integer/Real"));
      }else{
        term.set(type_bool);
        term.setRValue();
      }
    }
    term.set(term2.getLine(), term2.getPos());
    info.put(node, term);
  }

  public void outAUnsignedConstantFact(AUnsignedConstantFact node){
    Information fact = (Information)info.get(node.getUnsignedConstant());
    fact.setRValue();
    info.put(node, fact);
  }

  public void outAIdentFact(AIdentFact node){
    Information fact = new Information();
    Information ident = (Information)info.get(node.getIdent());
    try{
      Entry entry = table.lookup(ident.getString());
      if(!(entry instanceof hasType)){
        error_list.add(new SemanticError(ident.getLine(),
                                            ident.getPos(),
                                            ident+" can not use in expression"));
      }else{
        if(entry instanceof Constant || entry instanceof Variable
           || entry instanceof Parameter){
          fact.set(((hasType)entry).getType());
          fact.setRVariable();
        }else if(entry instanceof Function){
          if(((Function)entry).getNumberOfParameter() != 0){
            error_list.add(new SemanticError(ident.getLine(),
                                                ident.getPos(), 
                                                "Not match parameter"));
          }else{
            fact.set(((Function)entry).getType());
            fact.setRValue();
          }
        }
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }finally{
      fact.set(ident.getLine(),ident.getPos());
      info.put(node, fact);
    }
  }

  public void outAVariableTailFact(AVariableTailFact node){
    Information fact = new Information();
    Information ident = (Information)info.get(node.getIdent());
    Information variable_tail = (Information)info.get(node.getVariableTail());
    try{
      Type type = getVariableTailTypeR(ident, variable_tail.getList());
      if(type != null){
        fact.set(type);
        fact.setRVariable();
      }
    }catch(Exception e){
      error_list.add(e);
    }finally{    
      fact.set(ident.getLine(), ident.getPos());
      info.put(node, fact);
    }
  }

  public void outAFuncCallWithParamFact(AFuncCallWithParamFact node){
    Information fact = (Information)info.get(node.getFuncCallWithParam());
    fact.setRValue();
    info.put(node, fact);
  }

  public void outALparexprrparFact(ALparexprrparFact node){
    Information fact = (Information)info.get(node.getExpr());
    fact.setRValue();
    info.put(node, fact);
  }

  public void outANotfactFact(ANotfactFact node){
    Information fact = new Information();
    Information fact2 = (Information)info.get(node.getFact());
    if(fact2.getType() != type_bool){
      error_list.add(new SemanticError(fact2.getLine(),
                                          fact2.getPos(),
                                          "Operand is not Boolean"));
    }else{
      fact.set(type_bool);
      fact.setRValue();
    }
    fact.set(node.getNot().getLine(), node.getNot().getPos());    
    info.put(node, fact);
  }

  public void outABuiltinFuncCallFuncCallWithParam
                          (ABuiltinFuncCallFuncCallWithParam node){
    info.put(node, info.get(node.getBuiltinFuncCall()));
  }

  public void outAIdentFuncCallWithParam(AIdentFuncCallWithParam node){
    Information func_call = new Information();
    Information ident = (Information)info.get(node.getIdent());
    Information actual_param_list = (Information)info.get(node.getActualParamList());
    try{
      Entry entry = table.lookup(ident.getString());
      if(!(entry instanceof Function)){
        error_list.add(new SemanticError(ident.getLine(), 
                                            ident.getPos(), 
                                            ident.getString()+" is not Function"));
      }else{
        ConcreteAggregate actual_param = actual_param_list.getList();
        if(((Function)entry).getNumberOfParameter() != actual_param.getLength()){
          error_list.add(new SemanticError(ident.getLine(),
                                              ident.getPos(),
                                              "Not match parameter"));
        }else{
          Iterator formal_param_it = ((Function)entry).getParameter();
          Iterator actual_param_it = actual_param.iterator();
          while(actual_param_it.hasNext()){
            Information expr = (Information)actual_param_it.next();
            Parameter param = (Parameter)formal_param_it.next();
            if(expr.getType() != param.getType()){
              error_list.add(new SemanticError(expr.getLine(), 
                                                  expr.getPos(), 
                                                  "Not match parameter"));
            }
            if(param instanceof VariableParameter && (!expr.isRVariable())){
              error_list.add(new SemanticError(expr.getLine(), 
                                                  expr.getPos(), 
                                                  "Not match parameter"));
            }

          }
          func_call.set(((Function)entry).getType());
        }
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }finally{
      func_call.set(ident.getLine(),ident.getPos());
      info.put(node, func_call);
    }
  }

  public void outASinBuiltinFuncCall(ASinBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(!(type == type_int || type == type_real)){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(),
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_real);
    }
    func_call.set(node.getSin().getLine(), node.getSin().getPos());
    info.put(node, func_call);
  }

  public void outACosBuiltinFuncCall(ACosBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(!(type == type_int || type == type_real)){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(), 
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_real);
    }
    func_call.set(node.getCos().getLine(),node.getCos().getPos());
    info.put(node, func_call);
  }
  
  public void outAOddBuiltinFuncCall(AOddBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(type != type_int){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(), 
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_bool);
    }
    func_call.set(node.getOdd().getLine(), node.getOdd().getPos());
    info.put(node, func_call);
  }

  public void outAEvenBuiltinFuncCall(AEvenBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(type != type_int){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(), 
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_bool);
    }
    func_call.set(node.getEven().getLine(), node.getEven().getPos());
    info.put(node, func_call);
  }

  public void outASqrBuiltinFuncCall(ASqrBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(!(type == type_int || type == type_real)){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(), 
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type);
    }
    func_call.set(node.getSqr().getLine(), node.getSqr().getPos());
    info.put(node, func_call);
  }

  public void outAFArctanBuiltinFuncCall(AFArctanBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(!(type == type_int || type == type_real)){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(),
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_real);
    }
    func_call.set(node.getFArctan().getLine(),node.getFArctan().getPos());
    info.put(node, func_call);
  }

  public void outAFSqrtBuiltinFuncCall(AFSqrtBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(!(type == type_int || type == type_real)){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(),
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_real);
    }
    func_call.set(node.getFSqrt().getLine(),node.getFSqrt().getPos());
    info.put(node, func_call);
  }

  public void outAFAbsBuiltinFuncCall(AFAbsBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(!(type == type_int || type == type_real)){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(),
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type);
    }
    func_call.set(node.getFAbs().getLine(), node.getFAbs().getPos());
    info.put(node, func_call);
  }

  public void outAFRoundBuiltinFuncCall(AFRoundBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(type != type_real){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(),
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_int);
    }
    func_call.set(node.getFRound().getLine(), node.getFRound().getPos());
    info.put(node, func_call);
  }

  public void outAFTruncBuiltinFuncCall(AFTruncBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(type != type_real){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(),
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_int);
    }
    func_call.set(node.getFTrunc().getLine(), node.getFTrunc().getPos());
    info.put(node, func_call);
  }


  public void outFAExpBuiltinFuncCall(AFExpBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(!(type == type_int || type == type_real)){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(),
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_real);
    }
    func_call.set(node.getFExp().getLine(), node.getFExp().getPos());
    info.put(node, func_call);
  }

  public void outAFLnBuiltinFuncCall(AFLnBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(!(type == type_int || type == type_real)){
      error_list.add(new SemanticError(expr.getLine(), 
                                          expr.getPos(),
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_real); 
    }
    func_call.set(node.getFLn().getLine(), node.getFLn().getPos());
    info.put(node, func_call);
  }

  public void outAFChrBuiltinFuncCall(AFChrBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(type != type_int){
      error_list.add(new SemanticError(expr.getLine(),
                                          expr.getPos(), 
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_int);
    }
    func_call.set(node.getFChr().getLine(), node.getFChr().getPos());
    info.put(node, func_call);
  }

  public void outAFOrdBuiltinFuncCall(AFOrdBuiltinFuncCall node){
    Information func_call = new Information();
    Information expr = (Information)info.get(node.getExpr());
    Type type = expr.getType();
    if(!(type == type_int  || type == type_bool || type == type_char)){
      error_list.add(new SemanticError(expr.getLine(), 
                                          expr.getPos(), 
                                          "Mismatch Argument Type"));
    }else{
      func_call.set(type_int); 
    }
    func_call.set(node.getFOrd().getLine(), node.getFOrd().getPos());
    info.put(node, func_call);
  }

  public void outAFEofBuiltinFuncCall(AFEofBuiltinFuncCall node){
    Information func_call = new Information();
    func_call.set(type_bool);
    func_call.set(node.getFEof().getLine(), node.getFEof().getPos());
    info.put(node, func_call);
  }

  public void outAFEolnBuiltinFuncCall(AFEolnBuiltinFuncCall node){
    Information func_call = new Information();
    func_call.set(type_bool);
    func_call.set(node.getFEoln().getLine(), node.getFEoln().getPos());
    info.put(node, func_call);
  }

  public void outAActualParamList(AActualParamList node){
    info.put(node, info.get(node.getArgList()));
  }

  public void outAArgListArgList(AArgListArgList node){
    Information arg_list = (Information)info.get(node.getArgList());
    ConcreteAggregate list = arg_list.getList();
    list.add(info.get(node.getExpr()));
    info.put(node, arg_list);
  }

  public void outAExprArgList(AExprArgList node){
    Information arg_list = new Information();
    ConcreteAggregate list =  new ConcreteAggregate();
    list.add(info.get(node.getExpr()));
    arg_list.set(list);
    info.put(node, arg_list);
  }

  public void outASubscriptListVariableTail(ASubscriptListVariableTail node){
    Information subscript_list = (Information)info.get(node.getSubscriptList());
    ConcreteAggregate list = subscript_list.getList();
    if(node.getCustomTypeTail() instanceof AVariableTailCustomTypeTail){
      Information custom_type_tail
                = (Information)info.get(node.getCustomTypeTail());
      ConcreteAggregate list2 = custom_type_tail.getList();
      Iterator it =  list2.iterator();
      while(it.hasNext()){
        list.add(it.next());
      }
    }
    info.put(node, subscript_list);
  }

  public void outAFieldIdentVariableTail(AFieldIdentVariableTail node){
    Information var_tail = new Information();
    ConcreteAggregate list = new ConcreteAggregate();
    list.add(info.get(node.getIdent()));
    if(node.getCustomTypeTail() instanceof AVariableTailCustomTypeTail){
      Information custom_type_tail 
           = (Information)info.get(node.getCustomTypeTail());
      ConcreteAggregate list2 = custom_type_tail.getList();
      Iterator it =  list2.iterator();
      while(it.hasNext()){
        list.add(it.next());
      }
    }
    var_tail.set(list);
    info.put(node, var_tail);
  }

  public void outAVariableTailCustomTypeTail(AVariableTailCustomTypeTail node){
    info.put(node, info.get(node.getVariableTail()));
  }

  public void outASubscriptListSubscriptList(ASubscriptListSubscriptList node){
    Information subscript_list = (Information)info.get(node.getSubscriptList());
    ConcreteAggregate list =  subscript_list.getList();
    list.add(info.get(node.getExpr()));
    info.put(node, subscript_list);
  }

  public void outAExprSubscriptList(AExprSubscriptList node){
    Information sub_list = new Information();
    ConcreteAggregate list =  new ConcreteAggregate();
    list.add(info.get(node.getExpr()));
    sub_list.set(list);
    info.put(node, sub_list);
  }

  public void outANumberConstConstant(ANumberConstConstant node){
    info.put(node, info.get(node.getNumberConst()));
  }

  public void outAIdentConstant(AIdentConstant node){
    Information ident_constant = new Information();
    Information ident = (Information)info.get(node.getIdent());
    try{
      Entry entry =table.lookup(ident.getString());
      if(!(entry instanceof Constant)){
        error_list.add(new SemanticError(ident.getLine(),
                                            ident.getPos(),
                                            ident+" is not Constant"));
      }else{
        ident_constant.set(((Constant)entry).getValue());
        ident_constant.set(((Constant)entry).getType());
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
   }finally{
      ident_constant.set(ident.getLine(), ident.getPos());
      info.put(node, ident_constant);
    }
  }

  public void outAUnopConstant(AUnopConstant node){
    Information unop_constant = new Information();
    Information ident = (Information)info.get(node.getIdent());
    try{
      Entry entry =table.lookup(ident.getString());
      if(!(entry instanceof Constant)){
          error_list.add(new SemanticError(ident.getLine(), 
                                              ident.getPos(),
                                              ident + " is not Constant"));
      }else{
        Type type= ((Constant)entry).getType();
        if(!(type==type_int || type==type_real)){
          error_list.add(new SemanticError(ident.getLine(), 
                                              ident.getPos(),
                                              ident +" is not Integer/Real"));
        }else{
          String number = ((Constant)entry).getValue();
          if(node.getUnop() instanceof AMinusUnop){
	    number = "-" + number;
          }
          unop_constant.set(number);
          unop_constant.set(type);
        }
      }
    }catch(CompileError e){
      e.setLinePos(ident.getLine(), ident.getPos());
      error_list.add(e);
    }finally{
      unop_constant.set(ident.getLine(), ident.getPos());
      info.put(node, unop_constant);
    }
  } 

  public void outACharConstConstant(ACharConstConstant node){
    info.put(node, info.get(node.getCharConst()));
  }

  public void outAUnsignedNumberNumberConst(AUnsignedNumberNumberConst node){
    info.put(node, info.get(node.getUnsignedNumber()));
  }

  public void outAUnopNumberConst(AUnopNumberConst node){
    Information unsigned_number = (Information)info.get(node.getUnsignedNumber());
    if(node.getUnop() instanceof AMinusUnop){
      String number = "-"+ unsigned_number.getString();
      unsigned_number.set(number);
    }
    info.put(node, unsigned_number);
  }

  public void outAUnsignedNumber(AUnsignedNumber node){
    info.put(node, info.get(node.getNumber()));
  }

  public void outARealNumber(ARealNumber node){
    Information num = new Information();
    num.set(node.getReal().getText().trim());
    num.set(type_real);
    num.set(node.getReal().getLine(),  node.getReal().getPos());
    info.put(node, num);
  }

  public void outAIntegerNumber(AIntegerNumber node){
    Information num = new Information();
    num.set(node.getInteger().getText().trim());
    num.set(type_int);
    num.set(node.getInteger().getLine(),  node.getInteger().getPos());
    info.put(node, num);
  }

   
  public void outACharConst(ACharConst node){
    info.put(node, info.get(node.getCharacter()));
  }

  public void outACharacter(ACharacter node){
    Information char_info = new Information();
    String character = node.getChar().getText().trim();
    char_info.set(character);
    char_info.set(character.length()==3 ? type_char : type_string);
    char_info.set(node.getChar().getLine(), node.getChar().getPos());
    info.put(node, char_info);
  }

  public void outAUnsignedNumberUnsignedConstant
                                  (AUnsignedNumberUnsignedConstant node){
    info.put(node, info.get(node.getUnsignedNumber()));
  }

  public void outACharConstUnsignedConstant(ACharConstUnsignedConstant node){
    info.put(node, info.get(node.getCharConst()));
  }

  public void outABTrueUnsignedConstant(ABTrueUnsignedConstant node){
    Information constant = new Information();
    constant.set(node.getBTrue().getText().trim());
    constant.set(type_bool);
    constant.set(node.getBTrue().getLine(), node.getBTrue().getPos());
    info.put(node, constant);
  }

  public void outABFalseUnsignedConstant(ABFalseUnsignedConstant node){
    Information constant = new Information();
    constant.set(node.getBFalse().getText().trim());
    constant.set(type_bool);
    constant.set(node.getBFalse().getLine(), node.getBFalse().getPos());
    info.put(node, constant);
  }


//**********************************************

  public Type getVariableTailTypeR
           (Information ident, ConcreteAggregate variable_tail) throws Exception{
    try{
      Entry entry = table.lookup(ident.getString());
      if(! (entry instanceof Variable || entry instanceof Parameter)){
        throw new SemanticError(ident.getLine(), 
                                ident.getPos(), 
                                ident.getString() + " is not variable");
      }

      Type type=((hasType)entry).getType();
      Iterator it = variable_tail.iterator();
      while(it.hasNext()){
        if(type instanceof ArrayType){
          ConcreteAggregate index = ((ArrayType)type).getIndex();
          Iterator it2 = index.iterator();
          while(it2.hasNext()){
            Index dec = (Index)it2.next() ;
            Information use = (Information)it.next();
            if(use.getType() == null){ //tB[hꍇ
              throw new SemanticError(use.getLine(), 
                                      use.getPos(), 
                                      ident.getString()+" is not array");
            }
            if(dec.getType() != use.getType()){ //Y̌^Ⴄꍇ
              throw new SemanticError(use.getLine(),
                                      use.getPos(), 
                                      "Index type mismatch");
            }
          }
          type=((ArrayType)type).getType();
        }else if(type instanceof RecordType){ //R[h^̏ꍇ
          Information use = (Information)it.next();
          if(use.getString() == null){ //Y̏ꍇ
            throw new SemanticError(use.getLine(),
                                    use.getPos(), 
                                    ident.getString() + " is not record");
          }
          Entry entry2 = table.lookup(use.getString(), 
                                     ((RecordType)type).getField());
          if(! (entry2 instanceof Field)){ //tB[hł͂Ȃꍇ
            throw new SemanticError(use.getLine(), 
                                    use.getPos(),
                                    entry2.getName() + " is not field");
          }
          type = ((Field)entry2).getType();
        }else{ //v~eBu^̏ꍇ(gpꍇ)
            throw new SemanticError(ident.getLine(), 
                                    ident.getPos(),
                                    ident.getString() + " is not declared");
        }
      }
      if(type instanceof ArrayType || type instanceof RecordType){ 
                                           //gpȂꍇ
        throw new SemanticError(ident.getLine(),
                                ident.getPos(),
                                ident.getString() + " is not declared");
      }
      return type;
    }catch(CompileError e){
      if(e instanceof TableError){
        e.setLinePos(ident.getLine(), ident.getPos());
      }
      throw e;
    }
  }

  public Type getVariableTailTypeL
              (Information ident, ConcreteAggregate variable_tail, Entry proc)
              throws Exception{
    try{
      Entry entry = table.lookup(ident.getString());
      if(! (entry instanceof Variable || entry instanceof Parameter
            || entry instanceof Function)){
        throw new SemanticError(ident.getLine(), 
                                ident.getPos(),
                                ident.getString() + " is not variable");
      }
      if(entry instanceof Function && entry != proc){
          throw new SemanticError(ident.getLine(),
                                  ident.getPos(), 
                                  ident.getString() + " can not assign");
      }

      Type type = ((hasType)entry).getType();
      Iterator it = variable_tail.iterator();
      while(it.hasNext()){
        if(type instanceof ArrayType){
          ConcreteAggregate index = ((ArrayType)type).getIndex();
          Iterator it2 = index.iterator();
          while(it2.hasNext()){
            Index dec = (Index)it2.next() ;
            Information use = (Information)it.next();
            if(use.getType() == null){
              throw new SemanticError(use.getLine(),
                                      use.getPos(), 
                                      ident.getString()+" is not array");
            }
            if(dec.getType() != use.getType()){
              throw new SemanticError(use.getLine(),
                                      use.getPos(), 
                                      "Index type mismatch");
            }
          }
          type=((ArrayType)type).getType();
        }else if(type instanceof RecordType){
          Information use = (Information)it.next();
          if(use.getString() == null){
            throw new SemanticError(use.getLine(), 
                                    use.getPos(),
                                    ident.getString() + " is not record");
          }
          Entry entry2 = table.lookup(use.getString(), 
                                     ((RecordType)type).getField());
          if(! (entry2 instanceof Field)){
            throw new SemanticError(use.getLine(), 
                                    use.getPos(),
                                    entry2.getName() + " is not field");
          }
          type = ((Field)entry2).getType();
        }else{
            throw new SemanticError(ident.getLine(), 
                                    ident.getPos(),
                                    ident.getString() + " is not declared");
        }
      }
      if(type instanceof ArrayType || type instanceof RecordType){
        throw new SemanticError(ident.getLine(),
                                ident.getPos(), 
                                ident.getString() + " is not declared");
      }
      return type;
    }catch(CompileError e){
      if(e instanceof TableError){ 
        e.setLinePos(ident.getLine(), ident.getPos());
      }
      throw e;
    }
  }
}

class SemanticError extends CompileError{
  SemanticError(int line, int pos, String message){
    super(line, pos, message);
  }

  SemanticError(int line, String message){
    super(line, -1, message);
  }
}
